//
//  DCALcore.h
//  dcal
//
//  Created by Albert Tra on 31.03.17.
//  Copyright © 2017 DATEV eG. All rights reserved.
//

#ifndef DCALcore_h
#define DCALcore_h

#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include <stdlib.h>

#ifdef __cplusplus
extern "C" {
#endif
    // MARK: - Data structure
    typedef enum _RESULT_ERROR {
        ERROR_NONE = 0x0,
        ERROR_BAD_PARAMS = 0x1,
        ERROR_INIT = 0x2,
        ERROR_MEMORY = 0x3,
        ERROR_READER = 0x4,
        ERROR_SMARTCARD = 0x5,
        ERROR_CRYPTO = 0x6,
        ERROR_NETWORK = 0x7,
        ERROR_HTTP = 0x8,
        ERROR_PROXY = 0x9,
        ERROR_IO = 0xA,
        ERROR_CHACHE_MISS = 0xB,
        ERROR_FORBIDDEN = 0xC,
        ERROR_UNAUTHORIZED = 0xD,
        ERROR_UNSUPPORTED_REDIRECT = 0xE,
        
        
        // User triggered errors
        ERROR_USER_CANCELLED = 0x10,
        ERROR_USER_INPUT = 0x11,
        
        // Anything else
        ERROR_UNKNOWN = 0xFF,
        
    } RESULT_ERROR;
    
    typedef enum _LIBRARY {
        LIB_UNKNOWN = 0x0,
        LIB_SC_API = 0x1,
        LIB_SC_NET = 0x2,
        LIB_SC_PCSCS = 0x3,
        
        LIB_SC_API_DATEV = 0x10,
        LIB_SC_API_KOBIL = 0x11,
        LIB_SC_API_P12 = 0x1F,
        
        LIB_OPEN_SSL = 0x20,
        LIB_CURL = 0x21,
        
        LIB_SC_PCSC_MIDAIR = 0x30,
        LIB_SC_PCSC_KOBIL = 0x31,
        LIB_SC_PCSC_WINSCARD = 0x32,
        LIB_SC_PCSC_FEITIAN = 0x33,
        LIB_SC_PCSC_P12 = 0x3F,
        
        LIB_HTTP = 0x44, // Special case: HTTP Errors are not really a library but they map well to our responses anyway. Cause will contain the HTTP response code.
        
        LIB_PLATFORM_IOS = 0xA0,
        
        // Here start non-core libs.
        LIB_CoreCrypto = 0xB0,
    } LIBRARY;
    
    
    typedef struct _RESULT {
        int16_t cause : 16;
        LIBRARY lib : 8;
        RESULT_ERROR error : 8;
    } RESULT;
    
    typedef struct _MemoryStruct {
        char *memory;
        size_t size;
    } MemoryStruct;
    
    typedef struct _HTTP_REQUEST HTTP_REQUEST;
    
    typedef struct _HTTP_HEADER_FIELD {
        char* key;
        char* value;
        //if we trim whitespaces we need to keep original pointers
        char* headerlineptr;
        struct _HTTP_HEADER_FIELD *next;
    } HTTP_HEADER;
    
    typedef struct _HTTP_REQUEST HTTP_REQUEST;
    
    // Returning false will stop download.
    typedef bool (*progress_callback) (HTTP_REQUEST *request);
    typedef bool (*finished_callback)(HTTP_REQUEST *request);
    typedef bool (*error_callback)(char *url, int error, void *data);
    
    typedef void (*CallbackData) (void *);
    
    typedef struct _HTTP_REQUEST {
        //CURL *curl;
        //READER reader;
        
        char *url;
        char *method;
        
        int response_code;
        char *cookieFile;
        int cancel_request;
        
        ///
        size_t response_loaded;
        size_t response_loaded_before;
        size_t response_len;
        size_t response_last_start;
        
        /// Used to report transfer progress
        size_t totalBytesExpectedToSend;
        size_t totalBytesSent;
        size_t totalBytesExpectedToReceive;
        size_t didReceive;
        
        
        char *request_payload;
        char *response_payload;
        char *err_redirect_url;
        
        size_t request_loaded;
        size_t request_len;
        
        progress_callback on_progress;
        error_callback on_error;
        finished_callback on_finished;
        
        HTTP_HEADER *request_header;
        HTTP_HEADER *response_header;
        
        //DATEV_PROXY proxy;
        
        void *callback_data;
    } HTTP_REQUEST;
    
    // MARK: - Core Network
    
#ifdef ANDROID
    bool http_init();
    
#else
    _Bool http_init();
#endif
    uint32_t result_getIdentifier(RESULT result);
    
    // Returns a german string to the given error. The String must not be freed.
    const char *result_getString(RESULT result);
    
    void result_log(RESULT result);
    
    RESULT result_createSuccess(LIBRARY lib);
    
    RESULT reosult_create(LIBRARY lib, RESULT_ERROR error, int cause);
    
    char *result_getLibraryName(RESULT result);
    
    
    
    HTTP_REQUEST *_httpCreateRequest(char *url, char *method, unsigned char *payload, int payloadLen, progress_callback on_progress, error_callback on_error, finished_callback on_finished, void *callback_data);
    
    void http_cancel_request(HTTP_REQUEST *request);
    void http_destroy_request(HTTP_REQUEST *request);
    
    bool _httpAddHeaderField(HTTP_REQUEST *request, char *key, char *value);
    
    RESULT _httpPerformRequest(HTTP_REQUEST *request);
    // MARK: -
    MemoryStruct _get(const char* _endPoint, const char* _header);
    MemoryStruct _put(const char* _endPoint, const char* _header, const char* data, size_t size);
    
    // MARK: - OpenID Connect
    MemoryStruct _authorizationRequest(
                                       const char *redirectURI,
                                       const char *clientID,
                                       const char *scope,
                                       const bool isPoduction);
    MemoryStruct _requestToken(
                               const char *authorizationCode,
                               const char *redirectURI,
                               const char* cridential);
    /// Const pointer: itx | 197C8173-DDCE-457B-B6B0-FC4214319265
    MemoryStruct _refreshToken(
                               const char* cridential,
                               const char *refreshToken,
                               const bool isPoduction);
    MemoryStruct _revokeToken(
                              const char *token,
                              const char *refreshToken,
                              const char* credential,
                              const bool isProduction);
    
    size_t writeFunction(void *contents, size_t size, size_t nmemb, void *userp); /// Is used in DPCE
    
    // MARK: - Core Crypto
    MemoryStruct _encrypt(unsigned char const* data, int dataLength);
    MemoryStruct _encryptWithPublicKey(char const* data, int dataLength, char const* publicKey);
    MemoryStruct _encryptWithCert(const unsigned char * data, int dataLength, const unsigned char * pubCert, int certLength);
    MemoryStruct _decrypt(const unsigned char * data, int dataLength);
    
    // MARK: TMS
    MemoryStruct _sign(const unsigned char * data, int dataLength);
    MemoryStruct _tmsSign(const unsigned char * data, int dataLength, const unsigned char * pfxData, int pfxDataLength, const char* passphrase);
    void _parseCertificate(const unsigned char* rawCert, int certLength, const char* passphrase);
    MemoryStruct _getCert(unsigned char* rawCert, int certLength, const char* _key);
    MemoryStruct _parseCertificate_v2(unsigned char* rawCert, unsigned int certLength, const char* _key);
    void _parseFakeCertificate(char const* fakeSigner, char const* output); /// Only for testing
    MemoryStruct printPublicKey_v2(char const* fakeSigner); /// Only for testing
    void _setConfig( /// Only for testing
                    bool isTestingWithFakeCertificate,   /// For testing using fake certificate
                    bool loadEncryptedContentFromFile,   /// For testing decryption
                    bool loadPlainContentFromFile,        /// For testing encrytion from text file
                    char const* dummyInput,              /// Plain text input text file
                    char const* dummyEncryptedOutput    /// Encryted content in text file
    );
    // MARK: Validation
    bool _validateCertificate(const unsigned char * pfxData, int pfxDataLength, const char* passphrase);
    
#ifdef __cplusplus
}
#endif

#endif /* DCALcore_h */

